defmodule FileStream.Storage.GCS do
  @moduledoc """
  A module to stream Google Cloud Storage backed files.
  """

  @behaviour FileStream.Storage

  @opts_schema KeywordValidator.schema!(
                 scheme: [is: {:in, ["http", "https"]}, required: true, default: "https"],
                 host: [is: :binary, required: true, default: "storage.googleapis.com"],
                 port: [is: :integer, required: true, default: 443],
                 token: [is: {:one_of, [{:fun, 0}, :binary]}, required: true],
                 retries: [is: :integer, required: true, default: 5],
                 checkout_timeout: [is: :timeout, required: false, default: 8_000],
                 connect_timeout: [is: :timeout, required: false, default: 8_000],
                 recv_timeout: [is: :timeout, required: false, default: 15_000]
               )
  @http_opts [:checkout_timeout, :connect_timeout, :recv_timeout]

  ################################
  # FileStream.Storage Callbacks
  ################################

  @impl FileStream.Storage
  @spec init(keyword) :: map()
  def init(opts) do
    opts = KeywordValidator.validate!(opts, @opts_schema)
    {http_opts, opts} = Keyword.split(opts, @http_opts)
    opts = Keyword.put(opts, :http_opts, http_opts)
    Enum.into(opts, %{})
  end

  @impl FileStream.Storage
  def into(stream) do
    {init_write(stream), &write_stream/2}
  end

  @impl FileStream.Storage
  def reduce(stream, acc, fun) do
    Stream.resource(
      fn -> init_read(stream) end,
      &read_stream/1,
      &exit_stream/1
    ).(acc, fun)
  end

  ################################
  # Private API
  ################################

  defp init_write(stream) do
    uri = build_uri(stream, :upload)
    request(stream, :post, uri, :stream)
  end

  defp init_read(stream) do
    uri = build_uri(stream, :download)
    request(stream, :get, uri, ~c"", async: :once)
  end

  defp exit_stream(stream) do
    FileStream.reset_meta(stream)
  end

  defp request(stream, method, uri, body_or_stream, opts \\ [], attempt \\ 1) do
    headers = build_headers(stream)
    retries = stream.config.retries
    http_opts = opts ++ stream.config.http_opts

    case :hackney.request(method, uri, headers, body_or_stream, http_opts) do
      {:ok, ref} ->
        FileStream.put_meta(stream, :ref, ref)

      {:error, :closed} when attempt < retries ->
        request(stream, method, uri, body_or_stream, opts, attempt + 1)

      {:error, :timeout} ->
        raise FileStream.ConnectionTimeoutError, uri: stream.uri

      {:error, :nxdomain} ->
        raise FileStream.NXDomainError, uri: stream.uri

      {:error, reason} ->
        reason = format_error(reason)
        raise FileStream.Error, uri: stream.uri, reason: reason
    end
  end

  defp read_stream(stream) do
    ref = FileStream.get_meta(stream, :ref)

    with {:error, reason} <- :hackney.stream_next(ref) do
      reason = format_error(reason)
      raise FileStream.Error, uri: stream.uri, reason: reason
    end

    receive do
      {:hackney_response, ^ref, {:status, 200, _}} ->
        {[], stream}

      {:hackney_response, ^ref, {:status, status, _}} when status in [401, 403] ->
        raise FileStream.PermissionError, uri: stream.uri

      {:hackney_response, ^ref, {:status, 404, _}} ->
        raise FileStream.NotFoundError, uri: stream.uri

      {:hackney_response, ^ref, {:status, status, _}} ->
        reason = format_error({:status, status})
        raise FileStream.Error, uri: stream.uri, reason: reason

      {:hackney_response, ^ref, {:headers, _headers}} ->
        {[], stream}

      {:hackney_response, ^ref, chunk} when is_binary(chunk) ->
        {[chunk], stream}

      {:hackney_response, ^ref, :done} ->
        {:halt, stream}

      {:tcp_closed, _} ->
        reason = format_error(:tcp_closed)
        raise FileStream.Error, uri: stream.uri, reason: reason
    end
  end

  defp write_stream(stream, command) do
    ref = FileStream.get_meta(stream, :ref)

    case command do
      {:cont, body} ->
        :ok = :hackney.send_body(ref, body)
        stream

      :done ->
        stream = response(stream, ref)
        exit_stream(stream)

      :halt ->
        :ok
    end
  end

  defp response(stream, ref) do
    case :hackney.start_response(ref) do
      {:ok, 200, _, _} ->
        stream

      {:ok, status, _headers, _body} when status in [401, 403] ->
        raise FileStream.PermissionError, uri: stream.uri

      {:ok, status, _, _} ->
        reason = format_error({:status, status})
        raise FileStream.Error, uri: stream.uri, reason: reason

      {:error, reason} ->
        reason = format_error(reason)
        raise FileStream.Error, uri: stream.uri, reason: reason
    end
  end

  defp build_headers(%{config: %{token: token}}) do
    token =
      case token do
        token when is_function(token) -> token.()
        token when is_binary(token) -> token
      end

    [{"Authorization", "Bearer " <> token}]
  end

  defp build_uri(stream, :upload) do
    {bucket, name, uri} = build_uri(stream)
    path = Path.join(["/upload/storage/v1/b", bucket, "o"])
    query = URI.encode_query(%{"uploadType" => "media", "name" => name})
    URI.to_string(%{uri | path: path, query: query})
  end

  defp build_uri(stream, :download) do
    {bucket, name, uri} = build_uri(stream)
    name = URI.encode_www_form(name)
    path = Path.join(["/download/storage/v1/b/", bucket, "o", name])
    query = URI.encode_query(%{"alt" => "media"})
    URI.to_string(%{uri | path: path, query: query})
  end

  defp build_uri(stream) do
    bucket = stream.uri.host || ""
    name = String.trim_leading(stream.uri.path || "", "/")

    uri = %URI{
      scheme: stream.config.scheme,
      host: stream.config.host,
      authority: stream.config.host,
      port: stream.config.port
    }

    {bucket, name, uri}
  end

  defp format_error(error) do
    case error do
      :nxdomain -> "Domain does not exist."
      :req_not_found -> "Stream already complete."
      :tcp_closed -> "TCP closed."
      {:status, status} -> "Request error - status #{status}."
    end
  end
end
