defmodule FileStream.Storage.S3 do
  @moduledoc """
  A module to stream AWS S3 backed files.
  """

  @behaviour FileStream.Storage

  import SweetXml, only: [sigil_x: 2]

  @opts_schema KeywordValidator.schema!(
                 access_key_id: [is: :binary, required: true],
                 host: [is: :binary, default: "s3.amazonaws.com"],
                 port: [is: :integer, default: 443, required: true],
                 region: [is: :binary, required: true],
                 retries: [is: :integer, required: true, default: 5],
                 scheme: [is: {:in, ["http", "https"]}, required: true, default: "https"],
                 secret_access_key: [is: :binary, required: true],
                 chunk_size: [is: :integer, default: 5 * 1024 * 1024, required: true],
                 checkout_timeout: [is: :timeout, required: false, default: 8_000],
                 connect_timeout: [is: :timeout, required: false, default: 8_000],
                 recv_timeout: [is: :timeout, required: false, default: 15_000]
               )
  @http_opts [:checkout_timeout, :connect_timeout, :recv_timeout]

  ################################
  # FileStream.Storage Callbacks
  ################################

  @impl FileStream.Storage
  def init(opts) do
    opts = KeywordValidator.validate!(opts, @opts_schema)
    {http_opts, opts} = Keyword.split(opts, @http_opts)
    opts = Keyword.put(opts, :http_opts, http_opts)
    Enum.into(opts, %{})
  end

  @impl FileStream.Storage
  def into(stream) do
    {init_write(stream), &write_stream/2}
  end

  @impl FileStream.Storage
  def reduce(stream, acc, fun) do
    Stream.resource(
      fn -> init_read(stream) end,
      &read_stream/1,
      &exit_stream/1
    ).(acc, fun)
  end

  ################################
  # Private API
  ################################

  defp init_write(stream) do
    stream = build_uri(stream)
    uri = build_uri(stream, :init)

    upload_id =
      stream
      |> request(:post, uri, "", [:with_body])
      |> parse_initiate_multipart_upload()

    stream
    |> FileStream.put_meta(:upload_id, upload_id)
    |> FileStream.put_meta(:part_number, 1)
    |> FileStream.put_meta(:chunk, [])
    |> FileStream.put_meta(:chunk_size, 0)
    |> FileStream.put_meta(:etags, [])
  end

  defp init_read(stream) do
    stream = build_uri(stream)
    uri = build_uri(stream, :read)
    {stream, _, _} = request(stream, :get, uri, "", async: :once)
    stream
  end

  defp build_uri(stream) do
    FileStream.put_meta(stream, :uri, %URI{
      scheme: stream.config.scheme,
      host: stream.config.host,
      port: stream.config.port,
      path: "/#{stream.uri.host}#{stream.uri.path}"
    })
  end

  defp build_uri(stream, :read) do
    to_string(FileStream.get_meta(stream, :uri))
  end

  defp build_uri(stream, :init) do
    uri = FileStream.get_meta(stream, :uri)
    to_string(%{uri | query: "uploads=1"})
  end

  defp build_uri(stream, :part) do
    upload_id = FileStream.get_meta(stream, :upload_id)
    part_number = FileStream.get_meta(stream, :part_number)
    uri = FileStream.get_meta(stream, :uri)
    to_string(%{uri | query: "partNumber=#{part_number}&uploadId=#{upload_id}"})
  end

  defp build_uri(stream, :final) do
    upload_id = FileStream.get_meta(stream, :upload_id)
    uri = FileStream.get_meta(stream, :uri)
    to_string(%{uri | query: "uploadId=#{upload_id}"})
  end

  defp write_stream(stream, command) do
    case command do
      {:cont, body} ->
        maybe_upload_part(stream, body)

      :done ->
        stream
        |> upload_part([])
        |> finalize_multipart_upload()
        |> exit_stream()

      :halt ->
        :ok
    end
  end

  defp maybe_upload_part(stream, body) do
    chunk_size = FileStream.get_meta(stream, :chunk_size)
    chunk_size = byte_size(body) + chunk_size
    stream = FileStream.put_meta(stream, :chunk_size, chunk_size)

    if chunk_size >= stream.config.chunk_size do
      upload_part(stream, body)
    else
      chunk = FileStream.get_meta(stream, :chunk)
      FileStream.put_meta(stream, :chunk, [chunk, body])
    end
  end

  defp upload_part(stream, body) do
    uri = build_uri(stream, :part)
    chunk = FileStream.get_meta(stream, :chunk)
    chunk = IO.iodata_to_binary([chunk, body])

    stream
    |> request(:put, uri, chunk, [:with_body])
    |> add_etag_to_stream_meta()
    |> FileStream.put_meta(:chunk, [])
    |> FileStream.put_meta(:chunk_size, 0)
  end

  defp read_stream(stream) do
    ref = FileStream.get_meta(stream, :ref)

    with {:error, reason} <- :hackney.stream_next(ref) do
      reason = format_error(reason)
      raise FileStream.Error, uri: stream.uri, reason: reason
    end

    receive do
      {:hackney_response, ^ref, {:status, 200, _}} ->
        {[], stream}

      {:hackney_response, ^ref, {:status, status, _}} when status in [401, 403] ->
        raise FileStream.PermissionError, uri: stream.uri

      {:hackney_response, ^ref, {:status, 404, _}} ->
        raise FileStream.NotFoundError, uri: stream.uri

      {:hackney_response, ^ref, {:status, status, _}} ->
        reason = format_error({:status, status})
        raise FileStream.Error, uri: stream.uri, reason: reason

      {:hackney_response, ^ref, {:headers, _headers}} ->
        {[], stream}

      {:hackney_response, ^ref, chunk} when is_binary(chunk) ->
        {[chunk], stream}

      {:hackney_response, ^ref, :done} ->
        {:halt, stream}

      {:tcp_closed, _} ->
        reason = format_error(:tcp_closed)
        raise FileStream.Error, uri: stream.uri, reason: reason
    end
  end

  # credo:disable-for-next-line Credo.Check.Refactor.CyclomaticComplexity
  defp request(stream, method, uri, body_or_stream, opts, attempt \\ 1) do
    upload_id = FileStream.get_meta(stream, :upload_id)
    headers = build_headers(stream, body_or_stream, uri, upload_id, method)
    http_opts = opts ++ stream.config.http_opts
    retries = stream.config.retries

    case :hackney.request(method, uri, headers, body_or_stream, http_opts) do
      {:ok, ref} ->
        stream = FileStream.put_meta(stream, :ref, ref)
        {stream, nil, nil}

      {:ok, 200, headers, body} ->
        {stream, headers, body}

      {:ok, status, _headers, _body} when status in [401, 403] ->
        raise FileStream.PermissionError, uri: stream.uri

      {:ok, 404, _headers, _body} ->
        raise FileStream.NotFoundError, uri: stream.uri

      {:ok, status, _headers, body} ->
        reason = format_error({:status, status, body})
        raise FileStream.Error, uri: stream.uri, reason: reason

      {:error, :closed} when attempt < retries ->
        request(stream, method, uri, body_or_stream, opts, attempt + 1)

      {:error, :timeout} ->
        raise FileStream.ConnectionTimeoutError, uri: stream.uri

      {:error, :nxdomain} ->
        raise FileStream.NXDomainError, uri: stream.uri

      {:error, reason} ->
        reason = format_error(reason)
        raise FileStream.Error, uri: stream.uri, reason: reason
    end
  end

  defp build_headers(stream, body, uri, is_init, method) do
    datetime = :calendar.universal_time()

    []
    |> put_host_header(stream)
    |> put_amz_date_header(datetime)
    |> maybe_put_content_length_header(is_init, body, stream, method)
    |> put_content_hash_header(body)
    |> put_auth_header(stream, body, method, uri, datetime)
  end

  defp maybe_put_content_length_header(headers, is_init, body, stream, method) do
    if is_init || method == :post do
      content_length = if body == "", do: "0", else: "#{stream.meta.chunk_size}"
      headers ++ [{"content-length", content_length}]
    else
      headers
    end
  end

  defp put_host_header(headers, stream) do
    host = stream.config.host
    port = stream.config.port
    host_header = if port == 443, do: host, else: host <> ":" <> to_string(port)
    headers ++ [{"host", host_header}]
  end

  defp put_amz_date_header(headers, datetime) do
    headers ++ [{"x-amz-date", amz_date(datetime)}]
  end

  defp put_content_hash_header(headers, body) do
    headers ++ [{"x-amz-content-sha256", hash_body(body)}]
  end

  defp put_auth_header(headers, stream, body, method, uri, datetime) do
    query =
      uri
      |> URI.parse()
      |> query_from_parsed_uri()
      |> canonical_query_params()

    signature = signature(stream, method, headers, body, query, datetime)

    [
      {"Authorization",
       [
         "AWS4-HMAC-SHA256 Credential=",
         generate_credential_v4("s3", stream.config, datetime),
         ",",
         "SignedHeaders=",
         signed_headers(headers),
         ",",
         "Signature=",
         signature
       ]
       |> IO.iodata_to_binary()}
      | headers
    ]
  end

  defp signature(stream, method, headers, body, query, datetime) do
    uri = FileStream.get_meta(stream, :uri)

    request =
      build_canonical_request(
        method,
        headers,
        body,
        uri.path,
        query
      )

    string_to_sign = string_to_sign(request, datetime, "s3", stream.config)
    generate_signature_v4("s3", datetime, stream.config, string_to_sign)
  end

  defp build_canonical_request(method, headers, body, path, query) do
    method = method |> method_string |> String.upcase()
    headers = headers |> canonical_headers

    header_string =
      Enum.map_join(headers, "\n", fn
        {k, v} -> "#{k}:#{remove_dup_spaces(to_string(v))}"
      end)

    signed_headers_list = Enum.map_join(headers, ";", &elem(&1, 0))
    payload = hash_body(body)

    [
      method,
      "\n",
      path,
      "\n",
      query,
      "\n",
      header_string,
      "\n",
      "\n",
      signed_headers_list,
      "\n",
      payload
    ]
    |> IO.iodata_to_binary()
  end

  defp canonical_headers(headers) do
    headers
    |> Enum.reduce([], fn
      {k, _v}, acc when k in ["x-amzn-trace-id", "X-Amzn-Trace-Id"] -> acc
      {k, v}, acc when is_binary(v) -> [{String.downcase(to_string(k)), String.trim(v)} | acc]
      {k, v}, acc -> [{String.downcase(to_string(k)), v} | acc]
    end)
    |> Enum.sort(fn {k1, _}, {k2, _} -> k1 < k2 end)
  end

  defp remove_dup_spaces(str), do: remove_dup_spaces(str, "")
  defp remove_dup_spaces(str, str), do: str

  defp remove_dup_spaces(str, _last),
    do: str |> String.replace("  ", " ") |> remove_dup_spaces(str)

  defp string_to_sign(request, datetime, service, config) do
    request = hash_body(request)

    """
    AWS4-HMAC-SHA256
    #{amz_date(datetime)}
    #{generate_credential_scope_v4(service, config, datetime)}
    #{request}
    """
    |> String.trim_trailing()
  end

  defp generate_credential_v4(service, config, datetime) do
    scope = generate_credential_scope_v4(service, config, datetime)
    "#{config[:access_key_id]}/#{scope}"
  end

  defp generate_credential_scope_v4(service, config, datetime) do
    "#{date(datetime)}/#{config[:region]}/#{service}/aws4_request"
  end

  defp date({date, _time}) do
    date |> quasi_iso_format
  end

  defp quasi_iso_format({y, m, d}) do
    [y, m, d]
    |> Enum.map(&Integer.to_string/1)
    |> Enum.map(&zero_pad/1)
  end

  defp zero_pad(<<_>> = val), do: "0" <> val
  defp zero_pad(val), do: val

  defp generate_signature_v4(service, datetime, config, string_to_sign) do
    service
    |> signing_key(datetime, config)
    |> hmac_sha256(string_to_sign)
    |> Base.encode16(case: :lower)
  end

  defp signing_key(service, datetime, config) do
    ["AWS4", config[:secret_access_key]]
    |> hmac_sha256(date(datetime))
    |> hmac_sha256(config[:region])
    |> hmac_sha256(service)
    |> hmac_sha256("aws4_request")
  end

  defp signed_headers(headers) do
    headers
    |> Enum.map(fn {k, _} -> String.downcase(k) end)
    |> Kernel.--(["x-amzn-trace-id"])
    |> Enum.sort(&(&1 < &2))
    |> Enum.join(";")
  end

  defp query_from_parsed_uri(%{query: nil}), do: []

  defp query_from_parsed_uri(%{query: query_string}) do
    query_string
    |> URI.decode_query()
    |> Enum.to_list()
  end

  defp canonical_query_params(params) do
    params
    |> Enum.sort(&compare_query_params/2)
    |> Enum.map_join("&", &pair/1)
  end

  defp hmac_sha256(key, data), do: :crypto.mac(:hmac, :sha256, key, data)

  defp hash_body(body) do
    :sha256
    |> :crypto.hash(body)
    |> Base.encode16(case: :lower)
  end

  defp compare_query_params({key, value1}, {key, value2}), do: value1 < value2
  defp compare_query_params({key_1, _}, {key_2, _}), do: key_1 < key_2

  defp amz_date({date, time}) do
    date = date |> quasi_iso_format
    time = time |> quasi_iso_format

    [date, "T", time, "Z"]
    |> IO.iodata_to_binary()
  end

  defp method_string(method) do
    method
    |> Atom.to_string()
    |> String.upcase()
  end

  defp pair({k, _}) when is_list(k) do
    raise ArgumentError, "encode_query/1 keys cannot be lists, got: #{inspect(k)}"
  end

  defp pair({_, v}) when is_list(v) do
    raise ArgumentError, "encode_query/1 values cannot be lists, got: #{inspect(v)}"
  end

  defp pair({k, v}) do
    URI.encode_www_form(Kernel.to_string(k)) <> "=" <> aws_encode_www_form(Kernel.to_string(v))
  end

  defp aws_encode_www_form(str) when is_binary(str) do
    import Bitwise

    for <<c <- str>>, into: "" do
      case URI.char_unreserved?(c) do
        true -> <<c>>
        false -> "%" <> hex(bsr(c, 4)) <> hex(band(c, 15))
      end
    end
  end

  defp hex(n) when n <= 9, do: <<n + ?0>>
  defp hex(n), do: <<n + ?A - 10>>

  defp parse_initiate_multipart_upload(response) do
    {_, _, xml} = response

    SweetXml.xpath(xml, ~x"//InitiateMultipartUploadResult", upload_id: ~x"./UploadId/text()"s).upload_id
  end

  defp finalize_multipart_upload(stream) do
    uri = build_uri(stream, :final)
    body = build_final_body(stream)

    {stream, _, _} =
      stream
      |> FileStream.put_meta(:chunk_size, byte_size(body))
      |> request(:post, uri, body, [:with_body])

    stream
  end

  defp build_final_body(stream) do
    parts =
      stream
      |> FileStream.get_meta(:etags)
      |> Enum.reverse()
      |> Enum.map(fn {part_number, etag} ->
        [
          "<Part>",
          "<PartNumber>",
          Integer.to_string(part_number),
          "</PartNumber>",
          "<ETag>",
          etag,
          "</ETag>",
          "</Part>"
        ]
      end)

    ["<CompleteMultipartUpload>", parts, "</CompleteMultipartUpload>"]
    |> IO.iodata_to_binary()
  end

  defp add_etag_to_stream_meta(response) do
    {stream, headers, _} = response

    {_, etag} =
      Enum.find(headers, fn {k, _v} ->
        String.downcase(k) == "etag"
      end)

    etags = FileStream.get_meta(stream, :etags)
    part_number = FileStream.get_meta(stream, :part_number)
    etags = [{part_number, etag} | etags]

    stream
    |> FileStream.put_meta(:etags, etags)
    |> FileStream.put_meta(:part_number, part_number + 1)
  end

  defp exit_stream(stream) do
    FileStream.reset_meta(stream)
  end

  defp format_error(error) do
    case error do
      :nxdomain ->
        "Domain does not exist."

      :req_not_found ->
        "Stream already complete."

      :tcp_closed ->
        "TCP closed."

      :timeout ->
        "Request timeout."

      {:status, _code, body} ->
        SweetXml.xpath(body, ~x"//Error/Message/text()"s)

      {:status, status} ->
        "Request error - status #{status}."
    end
  end
end
